/** @file
 * Interface to the S/Ident requester library.
 *
 * This interface is heavily based on the libident interface, with some
 * additions for the Kerberos S/Ident functionality.
 *
 * @author libident written by Peter Eriksson <pen@lysator.liu.se> and Pdr
 *   Emanuelsson <pell@lysator.liu.se>
 * @author S/Ident written by Booker Bense <bbense@stanford.edu>.
 */

#ifndef SIDENT_H
#define SIDENT_H 1

#include <netinet/in.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>

/* BEGIN_DECLS is used at the beginning of declarations so that C++
   compilers don't mangle their names.  END_DECLS is used at the end.  Use
   macros so that smart editors don't get confused by the block. */
#undef BEGIN_DECLS
#undef END_DECLS
#ifdef __cplusplus
# define BEGIN_DECLS    extern "C" {
# define END_DECLS      }
#else
# define BEGIN_DECLS    /* empty */
# define END_DECLS      /* empty */
#endif

/** The registered ident port. */
#ifndef IDPORT
# define IDPORT 113
#endif

/** The service name for the S/Ident protocol. */
#define IDENT_AUTH_KRB_SERVICE "ident"

/** Default delay for network reads. */
#define IDENT_READ_TIMEOUT 60

/** Buffer size sufficient to hold a complete request or response. */
#define IDENT_BUFFER_SIZE 4096

/** Maximum length of an authenticator. */
#define MAX_AUTH_LEN 4096

/** Minimum length of an authenticator. */
#define MIN_AUTH_LEN 80 

/** Maximum length of an identifier.  Should be the same as MAX_K_NAME_SZ in
 * krb.h, but we don't want to depend on krb.h. */
#define MAX_AUTH_NAME 123 

/** The number of authentication flags supported by KERBEROS_V4. */
#define IDENT_MAX_KRB_AUTH_FLAG 1

/** A 16-bit unsigned type. */
#ifndef IDENT_INT16
# define IDENT_INT16 unsigned short
#endif

/** Bit to determine if mutual authentication is needed. */
#define MUTUAL_AUTH_BIT 0x01 


BEGIN_DECLS


/* VARIABLES AND STRUCTURES */


/** Version of the S/Ident library being used. */
extern char id_version[];

/** Return codes used by library functions. */
enum id_codes {
    IDENT_AUTH_OKAY = 0,        /**< No error. */

    /** An authentication verification operation failed. */
    IDENT_AUTH_FAIL,

    /** Authentication mechanism not supported.
     *
     * None of the authentication mechanisms offered by the requester are
     * supported for authentication of the owner of the indicated
     * connection.
     */
    IDENT_AUTH_NOT_SUPPORTED,

    /** No authentication information available.
     *
     * Indicates that the authentication mechanism is supported for the owner
     * of the connection, but that no authentication information is available
     * for that user.
     */
    IDENT_USER_CANT_AUTH,

    /** One or more values in auth-resp-info are syntactically invalid. */
    IDENT_INVALID_RESP_INFO,

    /** One or more values in auth-req-info are syntactically invalid. */
    IDENT_INVALID_REQ_INFO,

    /** Connection not found.
     *
     * The connection specified by the port pair is not currently in use or
     * currently not owned by an identifiable entity.
     */
    IDENT_NO_USER,

    /** Cannot determine connection owner for an unknown reason.
     *
     * Any error not covered by other codes should return this error code
     * value.  Optionally, this code MAY be returned in lieu of any other
     * specific error code if, for example, the server desires to hide
     * information implied by the return of that error code, or for any other
     * reason.  If a server implements such a feature, it MUST be configurable
     * and it MUST default to returning the proper error message.
     *
     * This error code is also used if the remote system closed the connection
     * without returning any response.
     */
    IDENT_UNKNOWN_ERROR,

    /** Port improperly specified.
     *
     * Either the local or foreign port was improperly specified.  This should
     * be returned if either or both of the port ids were out of range (TCP
     * port numbers are from 1-65535), negative integers, reals or in any
     * fashion not recognized as a non-negative integer.
     */
    IDENT_INVALID_PORT,

    /** User information intentionally hidden.
     *
     * The server was able to identify the user of this port, but the
     * information was not returned at the request of the user.
     */
    IDENT_HIDDEN_USER,

    /** User refused to authenticate.
     *
     * Indicates that the authentication mechanism is supported for the owner
     * of the connection, but that no authentication information is available
     * for that user; this value is used if the responder has prompted the
     * user to authenticate.
     *
     * @deprecated Use #IDENT_CANT_AUTH instead.
     */
    IDENT_USER_WONT_AUTH,

    /* All error codes below are internal */
    IDENT_INTERNAL_ERR,         /**< Internal library error. */
    IDENT_NO_MUTUAL_AUTH,       /**< No mutual authentication given. */
    IDENT_MUTUAL_AUTH_FAIL,     /**< Mutual authentication failed. */
    IDENT_SYSTEM_ERROR,         /**< Some internal system error. */
    IDENT_FLAG_NOT_SUPPORTED,   /**< Authentication flag not supported. */
    IDENT_INVALID_FLAG_VALUE,   /**< Invalid authentication flag value. */
    IDENT_TIMEOUT,              /**< Timeout in library functions. */
    IDENT_MAX_ERROR             /**< Fencepost value. */
};

/** A table mapping enum id_codes codes to text strings. */
extern const char *ident_err_txt[];

/** A table listing all authentication flags supported by KERBEROS_V4. */
extern const char *ident_krb_auth_flags[];

/** Low-level struct underlying an open ident connection. */
typedef struct {
    int fd;
    char buf[IDENT_BUFFER_SIZE];
} ident_t;

/** High-level struct encapsulating all the protocol data. */
typedef struct {
    int resp_port;              /**< Responder port. */
    int req_port;               /**< Requester port (i.e. us). */
    char *identifier;           /**< Identification string. */
    char *opsys;                /**< Operating system. */
    char *charset;              /**< Character set (what did you expect?). */
    int result_code;            /**< Result code from internal routines. */
    time_t expires;             /**< Expiration of authenticating ticket. */
    char *principal;            /**< Principal in local realm. */
} IDENT;

/** Internal structure containing requester information. */
struct ident_auth_client_data {
    /** Authenticated identity. */
    char user_ident[MAX_AUTH_NAME];

    /** Authenticated identity in local realm, if available. */
    char user_principal[MAX_AUTH_NAME];

    /** Authentication data from the network. */
    char authenticate_data[MAX_AUTH_LEN];

    int requester_port;         /**< Port of connection on requester. */
    int responder_port;         /**< Port of connection on responder. */
    int fd;                     /**< File descriptor of connection. */
    struct timeval *timeout;    /**< I/O timeout in seconds. */
    time_t expires;             /**< Expiration of authenticating ticket. */

    /** Local IP address. */
    struct in_addr *local_addr;

    /** Remote IP address. */
    struct in_addr *remote_addr;

    /** Pointer to authentication methods. */
    struct ident_requester_auth *auth_struct;

    /** Pointer to method-specific data used for authentication. */
    void *auth_method_data;
};

/** Internal structure containing responder information. */
struct ident_auth_data {
    FILE *fp;                   /**< Input stream. */
    char *port_data;            /**< Points to port data part of auth_line. */
    char *cmd_data;             /**< Points to command part of line. */
    char *authenticate_data;    /**< Points to AUTHENTICATE part of line. */
    int uid;                    /**< UID of process. */
    int pid;                    /**< PID of process. */

    /** Pointer to authentication methods. */
    struct ident_responder_auth *auth_struct;

    int local_port;             /**< Local port of connection. */
    int remote_port;            /**< Remote port of connection. */
  
    /** Local IP address. */
    struct in_addr *local_addr;

    /** Remote IP address. */
    struct in_addr *remote_addr;

    /** Pointer to method-specific data used for authentication. */
    void *auth_method_data;
};

/** Internal structure detailing how to do requester authentication. */
struct ident_requester_auth {
    char *auth_method;          /**< Name of authentication mechanism. */
    char *response;             /**< Response from server. */
    void *auth_method_init;     /**< Type-specific data (e.g. srvtab name). */
    char *auth_flags;           /**< Char string of requester flags. */

    /** Start a responder->requester authentication. */
    char *(*start)(struct ident_auth_client_data *);

    /** Do an authentication protocol exchange. */
    int (*auth)(struct ident_auth_client_data *);

    /** Free an authentication state structure. */
    void (*free_state)(void *);

    /** Set an authentication flag value. */
    int (*set_flag)(char *flag, char *value);

    /** Get the value of an authentication flag. */
    int (*get_flag)(char *flag, char **value);
};

/** Internal structure detailing how to do responder authentication. */
struct ident_responder_auth {
    char *auth_method;		/**< Name of authentication mechanism. */

    /** Start a responder->requester authentication. */
    int (*start)(struct ident_auth_data *);

    /** Do an authentication protocol exchange. */
    int (*auth)(struct ident_auth_data *);

    /** Free an authentication state structure. */
    void (*free_state)(void *);
};


/* HIGH-LEVEL CALLS */


/** Performs an ident query to identify the user on the other end of @a fd.
 *
 * The simplest of the lookup calls.  Assumes that @a fd points to an open
 * network connection and does an ident callback to the system on the other
 * end, asking for the identity of the user opening that connection.  Returns
 * only the identity string in newly allocated memory.
 *
 * @param fd File descriptor of open network connection to authenticate.
 * @param timeout Timeout in seconds.  0 means block until TCP times out.
 * @return The user identity in newly allocated memory, or NULL if the
 *   protocol failed or the user could not be authenticated.
 */
extern char *ident_id(int fd, int timeout);

/** Performs a full ident query for the user on the other end of @a fd.
 *
 * Like ident_id(), but returns a full #IDENT struct rather than just the
 * identity information.  The struct is returned even in the event of an error
 * on the remote side, so the error can be diagnosed more completely.
 *
 * @param fd File descriptor of open network connection to authenticate.
 * @param timeout Timeout in seconds.  0 means block until TCP times out.
 * @return An #IDENT struct in newly allocated memory, or NULL on basic
 *   errors.  If NULL is returned, errno will generally be set to something
 *   appropriate.
 */
extern IDENT *ident_lookup(int fd, int timeout);

/** Performs a full ident query.
 *
 * Like ident_lookup(), but rather than just taking the file descriptor of an
 * open network connection, takes the local and remote address and local and
 * remote ports instead.  This is the most complete of the ident lookup
 * functions and can be used to authenticate other open connections on the
 * same system provided that the local and remote ports are known.
 *
 * @param laddr Local address.
 * @param raddr Remote address.
 * @param resp_port Port of the connection on the remote system.
 * @param req_port Port of the connection on the local system.
 * @param timeout Timeout in seconds.  0 means block until TCP times out.
 * @return An #IDENT struct in newly allocated memory, or NULL on basic
 *   errors.  If NULL is returned, errno will generally be set to something
 *   appropriate.
 */
extern IDENT *ident_query (struct in_addr *laddr, struct in_addr *raddr,
                           int resp_port, int req_port, int timeout);
/** Frees allocated ident data.
 *
 * Frees the #IDENT struct returned by ident_lookup() or ident_query(),
 * including all data contained in it.
 *
 * @param id #IDENT struct to free.
 */
extern void ident_free(IDENT *id);

/** Set the type of ident authentication to use.
 *
 * Sets the type of authentication to request of the remote system.  Currently
 * there are only two supported values, "KERBEROS_V4" and "TRIVIAL".  There
 * are placeholders in the code for other values, such as GSS-API, but they
 * are not implemented.  "TRIVIAL" is the same as the normal ident protocol.
 *
 * @param auth_type Type of authentication ("KERBEROS_V4" or "TRIVIAL").
 * @param auth_init_data Type-specific initialization data.  For KERBEROS_V4,
 *   this should be the full path to the srvtab file that contains the key for
 *   the ident.<i>hostname</i> Kerberos principal.  TRIVIAL needs no data.
 * @return #IDENT_AUTH_OKAY on success and #IDENT_AUTH_NOT_SUPPORTED on
 *   failure.
 */
extern int ident_set_authtype(char *auth_type, void *auth_init_data);

/** Set a flag in the authentication request.
 *
 * The S/Ident protocol supports flags that can be set in the initial
 * authentication request.  This call sets a flag that will be passed to
 * future ident requests.  ident_set_authtype() must be called before this
 * call can be used.
 *
 * Currently, the only supported flag is "USER-INTERACTION" with possible
 * values "YES" or "NO", indicating whether the S/Ident server should attempt,
 * when possible, to bring up a dialog box to authenticate the user if they
 * are not already authenticated.
 *
 * @param flag The flag to set.
 * @param value The value to set it to.
 * @return #IDENT_AUTH_OKAY on success, #IDENT_FLAG_NOT_SUPPORTED if the flag
 *   isn't recognized, and #IDENT_INVALID_FLAG_VALUE if the value given isn't
 *   one of the valid values for that flag.
 */
extern int ident_set_authflag(char *flag, char *value);

/** Get the value of a flag.
 *
 * Returns the value of a flag set by ident_set_authflag().  As with that
 * function, ident_set_authtype() must be called before this function.  The
 * value returned in @a value points to internal memory and should not be
 * manipulated.
 *
 * @param flag The flag to query.
 * @param value Where to store a pointer to its value.  Will be set to the
 *   value of the flag if the flag is set or to NULL if it is not.
 * @return #IDENT_AUTH_OKAY if the flag is valid, or #IDENT_FLAG_NOT_SUPPORTED
 *   if it is not a recognized flag.
 */
extern int ident_get_authflag(char *flag, char **value);

/** Interrogates the error string returned by a responder.
 *
 * Interogates the error string returned from the responder for a given flag
 * and corresponding value.  The two currently defined flags are CAPABILITIES
 * and AUTH-MECH.  See the RFC for the corresponding values.  The responder
 * uses the extended error codes to allow the requester to attempt to
 * reformulate its call in an attempt to complete a successful transaction.
 * Currently, this can be used to distinguish between responders capable of
 * user interaction and those not capable.  In the future this mechanism may
 * be used to negotiate authentication methods.
 *
 * @param keyword Flag to search the error string for.
 * @param value Value that this keyword should have.
 * @param ident The #IDENT structure representing the error response.
 * @return #IDENT_AUTH_OKAY if that flag/value pair is found and
 *   #IDENT_AUTH_FAIL if it is not.
 */
extern int ident_query_error(char *keyword, char *value, IDENT *ident);


/* LOW-LEVEL CALLS AND MACROS */


/** Return the file descriptor of an ident connection.
 *
 * A macro that takes an #ident_t handle and returns the actual socket file
 * descriptor used for the connection to the remote server.
 *
 * @param id An #ident_t struct created by id_open().
 * @return The underlying file descriptor.
 */
#define id_fileno(id) ((id)->fd)

/** Open a new ident connection.
 *
 * Opens a connection to the remote IDENT server referred to by @a faddr.  The
 * timeout is specified by @a timeout.  Returns a pointer to an ident_t datum,
 * which is an opaque structure to be used as future reference to the opened
 * connection.
 *
 * @param laddr Local address.
 * @param faddr Server address.
 * @param timeout How long to wait for a response.  NULL means wait
 *   indefinitely, while a pointer to a zero-valued timeval struct sets
 *   non-blocking I/O in the same way as for select(2).
 * @return A valid #ident_t on success or NULL on failure.
 */
extern ident_t *id_open(struct in_addr *laddr, struct in_addr *faddr,
                        struct timeval *timeout);

/** Initialize an #ident_auth_client_data struct.
 *
 * @param auth_data Ignored.
 * @return A newly allocated #ident_auth_client_data struct.
 */
extern struct ident_auth_client_data *
    id_auth_init(struct ident_auth_client_data *auth_data);

/** Close an open ident connection.
 *
 * Closes the connection opened with id_open() and frees all data associated
 * with @a id.
 *
 * @param id Connection to free.
 * @return 0 on success, non-zero on failure.  errno will be set on failure.
 */
extern int id_close(ident_t *id);

/** Send query to remote ident server.
 *
 * Sends off a query to a remote ident server.  @a resp_port and @a req_port
 * are sent to the server to identify the connection for which identification
 * is needed.  @a timeout is given as for id_open().
 *
 * @param id Pointer to open ident connection from id_open().
 * @param resp_port Port of the connection on the remote system.
 * @param req_port Port of the connection on the local system.
 * @param timeout Timeout as per id_open().
 * @param auth_data Authentication data struct specifying the authentication
 *   protocol to use.
 * @return Returns the number of bytes sent to the remote server on success,
 *   or -1 on failure.  On failure, errno is set.
 */
extern int id_query(ident_t *id, int resp_port, int req_port,
                    struct timeval *timeout,
                    struct ident_auth_client_data *auth_data);

/** Parse the reply to an ident query.
 *
 * Parses the reply to a query sent off by id_query() and returns information
 * to the locations pointed to by @a resp_port, @a req_port, @a identifier, @a
 * opsys, and @a charset.  For string data (@a identifier, @a opsys, and @a
 * charset) pointers to allocated space are returned.  The caller is
 * responsible for freeing.
 *
 * @param id Pointer to open ident connection from id_open().
 * @param resp_port Port of the connection on the remote system.
 * @param req_port Port of the connection on the local system.
 * @param identifier User identity.
 * @param opsys Operating system of ident server.
 * @param charset Character set of returned information.
 * @param auth_data Authentication data struct specifying the authentication
 *   protocol to use.
 * @return Returns 1 on success, and in general:
 *   <dl>
 *   <dt>-3</dt>
 *   <dd>Illegal reply type from remote server.  @a identifier is set to
 *       the illegal reply.</dd>
 *
 *   <dt>-2</dt>
 *   <dd>Cannot parse the reply from the server.  @a identifier is normally
 *       set to the illegal reply.</dd>
 *
 *   <dt>-1</dt>
 *   <dd>Some other general error or timeout.</dd>
 *
 *   <dt>0</dt>
 *   <dd>Non-blocking mode is set and id_parse() has not finished parsing the
 *       reply from the server.</dd>
 *
 *   <dt>1</dt>
 *   <dd>Complete success.</dd>
 *
 *   <dt>2</dt>
 *   <dd>The query and reply (the protocol) was successful, but the remote
 *       site experienced some error.  @a identifier is set to the error
 *       message from the remote server.</dd>
 *   </dl>
 */
extern int id_parse(ident_t *id, struct timeval *timeout,
                    int *resp_port, int *req_port,
                    char **identifier, char **opsys, char **charset,
                    struct ident_auth_client_data *auth_data);

END_DECLS

#endif /* SIDENT_H */
