/** @file
 * High-level interface to the ident library.
 *
 * @author Pdr Emanuelsson <pell@lysator.liu.se>
 * @author hacked by Peter Eriksson <pen@lysator.liu.se>
 *
 * $Id: ident.c,v 1.6 2003/07/24 19:08:19 eagle Exp $
 */

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>

/* BSDI needs <netinet/in.h> before <arpa/inet.h>. */
#include <netinet/in.h>
#include <arpa/inet.h>

#include "sident.h"

#define MAX_PORT 65536
#define VALID_PORT( _port_ ) ((_port_) > 1 && (_port_) < MAX_PORT )


/* Do a complete ident query and return result */
IDENT *
ident_lookup(int fd, int timeout)
{
    struct sockaddr_in localaddr, remoteaddr;
    int len;
    
    len = sizeof(remoteaddr);
    if (getpeername(fd, (struct sockaddr*) &remoteaddr, &len) < 0)
	return 0;
    
    len = sizeof(localaddr);
    if (getsockname(fd, (struct sockaddr *) &localaddr, &len) < 0)
	return 0;

    return ident_query( &localaddr.sin_addr, &remoteaddr.sin_addr,
			(int ) ntohs(remoteaddr.sin_port), 
			(int ) ntohs(localaddr.sin_port),
			timeout);
}


IDENT *
ident_query(struct in_addr *laddr, struct in_addr *raddr,
            int resp_port, int req_port, int timeout)
{
    int res;
    ident_t *id;
    struct timeval timout;
    IDENT *ident=NULL;
    struct ident_auth_client_data *auth_data ; 
    char tmp[IDENT_BUFFER_SIZE] ; 
    /* Do some sanity checks on the args */ 
    if ( (!( VALID_PORT(resp_port) && VALID_PORT(req_port)))
	 || timeout < 0 ) { 
      errno = EINVAL ; 
      return NULL ; 
    }

    timout.tv_sec = timeout;
    timout.tv_usec = 0 ;

    auth_data = id_auth_init(auth_data);
    auth_data->local_addr = laddr ; 
    auth_data->remote_addr = raddr ; 

    if (timeout)
	id = id_open( laddr, raddr, &timout);
    else
	id = id_open( laddr, raddr, (struct timeval *)NULL);
    
    if (!id)
    {
	errno = EINVAL;
	return NULL;
    }
  
    if (timeout)
	res = id_query(id, resp_port, req_port, &timout,auth_data);
    else
	res = id_query(id, resp_port, req_port, (struct timeval *) NULL,auth_data);
    
    if (res < 0)
    {
        id_auth_free(auth_data); 
	id_close(id);
	return NULL ;
    }
    
    ident = (IDENT *) malloc(sizeof(IDENT));
    if (!ident) {
        id_auth_free(auth_data); 
	id_close(id);
	return NULL;
    }
    
    if (timeout)
	res = id_parse(id, &timout,
		       &ident->resp_port,
		       &ident->req_port,
		       &ident->identifier,
		       &ident->opsys,
		       &ident->charset,
		       auth_data);
    else
	res = id_parse(id, (struct timeval *) 0,
		       &ident->resp_port,
		       &ident->req_port,
		       &ident->identifier,
		       &ident->opsys,
		       &ident->charset,
                       auth_data);

    /* Copy over the principal, if any, first, since this is another memory
       allocation that could potentially fail and change the result code. */
    if ( auth_data->user_principal != NULL ) {
      ident->principal = strdup(auth_data->user_principal) ;
      if ( ident->principal == NULL )
        res = IDENT_SYSTEM_ERROR ;
    }

    /* Check to see if we succeeded.  If we didn't, cobble up an error reply
       to stuff into the identifier field. */
    if ( res != IDENT_AUTH_OKAY ) { 
      if ( ident->identifier != NULL ) {
	sprintf(tmp,"%s:%s",ident_err_txt[res],ident->identifier);
      	free(ident->identifier);
      } else { 
	sprintf(tmp,"%s:NULL",ident_err_txt[res]); 
      }
      ident->identifier = strdup(tmp); 
    }

    /* Success.  Set the rest of the ident struct, free our internal
       structures, and return. */
    ident->result_code = res ;
    ident->expires = auth_data->expires ;
    id_auth_free(auth_data); 
    id_close(id);
    return ident;
}


char *
ident_id(int fd, int timeout)
{
    IDENT *ident;
    char *id=0;
    
    ident = ident_lookup(fd, timeout);
    if (ident && ident->identifier && *ident->identifier &&
	ident->result_code == IDENT_AUTH_OKAY )
    {
	id = strdup(ident->identifier);
	if (id == NULL)
	    return NULL;
    }

    ident_free(ident);
    return id;
}


void
ident_free(IDENT *id)
{
    if (!id)
	return;
    if (id->identifier)
	free(id->identifier);
    if (id->opsys)
	free(id->opsys);
    if (id->charset)
	free(id->charset);
    free(id);
}


