/** @file
 * SASL-based authentication mechanisms for ident.
 *
 * @author Booker Bense <bbense@stanford.edu>
 * @author GSS-API support added by Russ Allbery <rra@stanford.edu>
 * @date 1996
 *
 * $Id: ident_auth.c,v 1.9 2003/07/24 19:08:19 eagle Exp $
 */

#include "config.h"

#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h> 
#include <signal.h>
#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <sys/param.h>
#include <sys/socket.h> 
#include <sys/time.h>
#include <sys/types.h>

#include <krb.h>

#include "sident.h"
 
/* Exported definition of requester authentication mechanism */
static char *krb_requester_start(); 
static char *null_requester_start(); 
static int krb_requester_auth(); 
static int null_requester_auth(); 
static void krb_free_state(); 
static void null_free_state(); 
static char *trivial_requester_start(); 
static int trivial_requester_auth();
static int krb_set_flag(); 
static int krb_get_flag(); 
static int null_set_flag(); 
static int null_get_flag();

/* GSS-API routines. */
extern char *gssapi_requester_start();
extern int gssapi_requester_auth();
extern void gssapi_free_state();
 
extern char *sfgets() ; 

struct ident_requester_auth krb_ident_auth = {
    "KERBEROS_V4",
    NULL,
    NULL, 
    NULL,
    krb_requester_start,
    krb_requester_auth,
    krb_free_state,
    krb_set_flag,
    krb_get_flag
};

struct ident_requester_auth gssapi_ident_auth = {
    "GSSAPI",
    NULL,
    NULL,
    NULL,
    gssapi_requester_start,
    gssapi_requester_auth,
    gssapi_free_state,
    krb_set_flag,
    krb_get_flag
};

struct ident_requester_auth otp_ident_auth = {
    "OTP_RFC1938",
    NULL,
    NULL,
    NULL,
    null_requester_start,
    null_requester_auth,
    null_free_state,
    null_set_flag,
    null_get_flag
};

struct ident_requester_auth trivial_ident_auth = {
    "TRIVIAL",
    NULL,
    NULL,
    NULL, 
    trivial_requester_start,
    trivial_requester_auth,
    null_free_state,
    null_set_flag,
    null_get_flag
};


#define IDENT_METHODS_COUNT 4
/* TRIVIAL method must always be the last method */ 
#define IDENT_TRIVIAL_METHOD (IDENT_METHODS_COUNT - 1 ) 
struct ident_requester_auth *auth_methods[IDENT_METHODS_COUNT] = { &krb_ident_auth, 
						&gssapi_ident_auth,
						&otp_ident_auth,
						&trivial_ident_auth
}; 

/* Start with Trivial as the default */ 
struct ident_requester_auth *global_auth_method = &trivial_ident_auth ; 


int 
ident_set_authtype( char *auth_type , void * auth_init_data ) 
{ 
  char *t ; 
  int i ;
  
  for ( i = 0 ; i < IDENT_METHODS_COUNT ; i++ ) { 
    t = auth_methods[i]->auth_method ; 
    if ( strncmp(auth_type,t,strlen(t) ) == 0 ) {
      global_auth_method = auth_methods[i] ;
    
      auth_methods[i]->auth_method_init = auth_init_data ; 
      return IDENT_AUTH_OKAY ;
    }  
  } 

    return IDENT_AUTH_NOT_SUPPORTED ; 
    
}

int 
ident_set_authflag( char *flag , char *value ) {
 
  return global_auth_method->set_flag( flag, value) ; 

  
} 

int 
ident_get_authflag( char *flag , char **value ) { 

  return global_auth_method->get_flag( flag, value) ; 
  
} 

int 
ident_query_error( char *keyword, char *value, IDENT *ident ) { 
  
  char *e_string,*extension,*tmp ; 
  if ( ident->result_code == IDENT_AUTH_OKAY  ) { 
    return IDENT_AUTH_FAIL ; 
  }

  e_string = ident->identifier ; 
  /* Format is error_code_string:everything in response after ERROR */ 
  if ( e_string == NULL ) { 
    return IDENT_AUTH_FAIL ; 
  }
  tmp = strchr(e_string,':') ; 
  if ( tmp == NULL ) { 
        return IDENT_AUTH_FAIL ; 
  }
  extension = strchr(tmp,':') ; 
  if ( extension == NULL ) { 
        return IDENT_AUTH_FAIL ; 
  }
  tmp = strstr(extension,keyword); 
  if ( tmp == NULL ) { 
        return IDENT_AUTH_FAIL ; 
  }
  e_string = tmp ; 
  tmp += strlen(keyword) ; 
  if ( *tmp != '=' ) { 
        return IDENT_AUTH_FAIL ; 
  }
  e_string = ++tmp ; 
  if( strstr(e_string,value) == NULL ) { 
        return IDENT_AUTH_FAIL ; 
  }
  return IDENT_AUTH_OKAY ; 
} 

struct ident_auth_client_data 
*id_auth_init(struct ident_auth_client_data *auth_data)
{ 
  auth_data = (struct ident_auth_client_data *) malloc(sizeof(struct ident_auth_client_data)); 

  auth_data->user_ident[0] = '\0';
  auth_data->user_principal[0] = '\0';
  auth_data->auth_struct = global_auth_method ; 

  return auth_data ; 
} 

void
id_auth_free(struct ident_auth_client_data *auth_data)
{ 
  /* free kstate stuff */ 
  free(auth_data->auth_method_data); 
  /* free the auth_data struct */ 
  free(auth_data); 
} 

int 
parse_auth(line,auth_struct)
char *line;  
struct ident_requester_auth **auth_struct ; 
{ 
  int rcode ;
  int i ; 
  char auth_method[128] ;
  *auth_struct = NULL ; 
  if ( line == NULL ) { 
    *auth_struct = auth_methods[IDENT_TRIVIAL_METHOD] ; 
    return IDENT_AUTH_OKAY ; 
  }
  /* Grok to the next , */ 

  rcode = sscanf(line ," %32[^ \t\n\r,]",auth_method ); 
  

  for(i = 0 ; i < IDENT_METHODS_COUNT ; i++ ) { 
    if ( strncmp(auth_method,auth_methods[i]->auth_method,
		 strlen(auth_methods[i]->auth_method)) == 0 ) {
      *auth_struct = auth_methods[i] ; 
      return(IDENT_AUTH_OKAY) ; 
    }
  }
  return(IDENT_INVALID_REQ_INFO); 
  
  
} 


/* All the code from here to EOF is C&P from imclient.c in the imap.1.4
libraries. 
*/ 

#define XX 127
/*
 * Tables for encoding/decoding base64
 */
static const char basis_64[] =
   "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const char index_64[256] = {
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,62, XX,XX,XX,63,
    52,53,54,55, 56,57,58,59, 60,61,XX,XX, XX,XX,XX,XX,
    XX, 0, 1, 2,  3, 4, 5, 6,  7, 8, 9,10, 11,12,13,14,
    15,16,17,18, 19,20,21,22, 23,24,25,XX, XX,XX,XX,XX,
    XX,26,27,28, 29,30,31,32, 33,34,35,36, 37,38,39,40,
    41,42,43,44, 45,46,47,48, 49,50,51,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
};
#define CHAR64(c)  (index_64[(unsigned char)(c)])

/*
 * Decode in-place the base64 data in 'input'.  Returns the length
 * of the decoded data, or -1 if there was an error.
 */
int 
ident_decodebase64(input)
char *input;
{
    int len = 0;
    unsigned char *output = (unsigned char *)input;
    int c1, c2, c3, c4;

    while (*input) {
	c1 = *input++;
	if (CHAR64(c1) == XX) return -1;
	c2 = *input++;
	if (CHAR64(c2) == XX) return -1;
	c3 = *input++;
	if (c3 != '=' && CHAR64(c3) == XX) return -1; 
	c4 = *input++;
	if (c4 != '=' && CHAR64(c4) == XX) return -1;
	*output++ = (CHAR64(c1) << 2) | (CHAR64(c2) >> 4);
	++len;
	if (c3 == '=') break;
	*output++ = ((CHAR64(c2) << 4) & 0xf0) | (CHAR64(c3) >> 2);
	++len;
	if (c4 == '=') break;
	*output++ = ((CHAR64(c3) << 6) & 0xc0) | CHAR64(c4);
	++len;
    }

    return len;
}
	

/*
 * Write to the string 'output' the base-64 encoded data
 * 'input', of (unencoded) length 'len'.
 */

void
ident_writebase64(input, output, len)
char *input;
char *output;
int len;
{
    char buf[IDENT_BUFFER_SIZE];
    char *outptr ; 
    int buflen = 0;
    int c1, c2, c3;
    outptr = output ; 
    while (len) {
	if (buflen >= sizeof(buf)-4) {
	    strncpy(outptr,buf,buflen);
	    buflen = 0;
	    outptr += buflen ; 
	}
	
	c1 = (unsigned char)*input++;
	buf[buflen++] = basis_64[c1>>2];

	if (--len == 0) c2 = 0;
	else c2 = (unsigned char)*input++;
	buf[buflen++] = basis_64[((c1 & 0x3)<< 4) | ((c2 & 0xF0) >> 4)];

	if (len == 0) {
	    buf[buflen++] = '=';
	    buf[buflen++] = '=';
	    break;
	}

	if (--len == 0) c3 = 0;
	else c3 = (unsigned char)*input++;

	buf[buflen++] = basis_64[((c2 & 0xF) << 2) | ((c3 & 0xC0) >>6)];
	if (len == 0) {
	    buf[buflen++] = '=';
	    break;
	}

	--len;
	buf[buflen++] = basis_64[c3 & 0x3F];
    }

    if (buflen >= sizeof(buf)-2) {
	strncpy(outptr, buf, buflen);
	outptr += buflen ; 
	buflen = 0;
    }
    buf[buflen++] = '\0';
    
    strncpy(outptr, buf, buflen);
}

/* 
**
**     Kerberos auth routines
**
**
*/ 


/* acte_krb.c -- KERBEROS_V4 authentication routines for IMAP.
 *
 *	(C) Copyright 1994 by Carnegie Mellon University
 *
 *                      All Rights Reserved
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of Carnegie
 * Mellon University not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  Carnegie Mellon University makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 */


extern char *lcase();
extern char *krb_get_phost(), *krb_realmofhost();

static char *srvtab = "";	/* Srvtab filename */

/* Maximum number of bytes of overhead the protection mechanisms use */
#define PROTECTION_OVERHEAD 31

int
do_out(char *string,  struct ident_auth_client_data *auth_data ) 
{ 
    RETSIGTYPE (*old_sig)(int);
    int res;
    char buf[IDENT_BUFFER_SIZE];
    fd_set ws;
 
  
  if ( auth_data != NULL ) { 
      sprintf(buf, "%d , %d : AUTHENTICATE : %s\r\n\0",
	      
	      auth_data->responder_port,
	      auth_data->requester_port,
	      string); 
  } else { 
    return -1 ; 
  }

   
#ifdef DEBUG 
    printf("Sending.. \n%s",buf) ; 
#endif 
    if (auth_data->timeout)
    {
	FD_ZERO(&ws);
	FD_SET(auth_data->fd, &ws);

	if ((res = select(FD_SETSIZE, NULL, &ws, NULL, auth_data->timeout)) < 0)
	    return -1;
	
	if (res == 0)
	{
	    errno = ETIMEDOUT;
	    return -1;
	}
    }

    old_sig = signal(SIGPIPE, SIG_IGN);
    
    res = write(auth_data->fd, buf, strlen(buf));
    
    signal(SIGPIPE, old_sig);
    
    return res;


} 

/* Private state used by this mechanism */
struct krb_state {
    /* common */
    char service[MAX_K_NAME_SZ+1];
    int authstepno;
    des_cblock session;	/* Our session key */
    des_key_schedule schedule; /* Schedule for our session key */
    long challenge;
    char user[MAX_K_NAME_SZ+1];
    int protallowed;
    int maxbufsize;
    struct sockaddr_in localaddr, remoteaddr;
    long prot_time_sec;
    char prot_time_5ms;
    /* client */
    char instance[INST_SZ];
    char realm[REALM_SZ];
    char string[2048] ; /* store built responses */ 
    /* server */
    int (*authproc)();
    AUTH_DAT kdata;
};

/*
 * Free the space used by an opaque state pointer
 */
static void
krb_free_state(state)
void *state;
{
    memset((char *)state, 0, sizeof(struct krb_state));
    free((char *) state);
}

/* 
 * Flag handling routines : should I spend the time to do this right? 
                            Will there ever be more than a single stupid flag? 
 */ 
static int 
krb_set_flag(char *flag, char *value ) { 
  int i ; 
  if ( flag != NULL && value != NULL ) { 
    for ( i = 0 ; i < IDENT_MAX_KRB_AUTH_FLAG ; i++ ) {
      if ( strncmp(flag,ident_krb_auth_flags[i],
		   strlen(ident_krb_auth_flags[i])) == 0 ) { 
	/* Check if value is okay */ 
	/* This is a hosing cheat, but at least it will be possible to 
	 * do  it right if we ever decide we need YAAF.  
	 */ 
	if ( strncmp(value,"YES",3 ) == 0 ) { 
	  global_auth_method->auth_flags = "USER-INTERACTION=YES" ; 
	}else if (strncmp(value,"NO",2) == 0 ) { 
	  global_auth_method->auth_flags = "USER-INTERACTION=NO" ; 
	}else { 
	  return IDENT_INVALID_FLAG_VALUE ;
	} 
	return IDENT_AUTH_OKAY ; 
      } 
    }
  }
    
  return IDENT_FLAG_NOT_SUPPORTED ; 
} 

static int 
krb_get_flag(char *flag, char **value ) { 
 int i ; 
 char *ptr_flag ; 
 char *ret_ptr ; 

 *value = NULL ; 
 

 if ( flag != NULL ) { 
   for ( i = 0 ; i < IDENT_MAX_KRB_AUTH_FLAG ; i++ ) { 
     if ( strncmp(flag,ident_krb_auth_flags[i],
		  strlen(ident_krb_auth_flags[i])) == 0 ) { 
       ptr_flag = strstr(global_auth_method->auth_flags,ident_krb_auth_flags[i]) ; 
       if ( ptr_flag != NULL ) {   
	 /* Check if value is okay */
	 ret_ptr = strchr(ptr_flag,'=' ) ; /* This breaks when we have 
					  *  multiple flags in auth_flags.
					  *  It does? 
					  */ 
	 if (ret_ptr != NULL ) {
	   (ret_ptr)++ ; 
	   *value = ret_ptr ; 
	 }
       } 
       return IDENT_AUTH_OKAY ; /* Flag is valid, value maybe null */ 
     }
   }
 }
 /* Not a flag we recognize */  
 return IDENT_FLAG_NOT_SUPPORTED ; 
} 



/*
 * Start the requester side of an authentication exchange.
 */
static char *krb_requester_start(auth_data)
  struct ident_auth_client_data *auth_data ; 
{

    struct krb_state *kstate;
    char *p ; 
    char tmp[2048] ;
    int32_t k;
    kstate = ( struct krb_state *) malloc(sizeof(struct krb_state)) ; 
    if ( kstate == NULL )
      return NULL  ; 
    /* set up service */  

    strcpy(kstate->service,IDENT_AUTH_KRB_SERVICE) ; 
    
    /* Just need to return a string that has an auth type */ 
    /* i.e. KERBEROS_V4, 444444== */ 
    kstate->authstepno = 0 ; 
    p = kstate->string ; 
    kstate->challenge = rand() ^ time(0) ^ getpid()  ;
#ifdef DEBUG
    printf("Challenge is %ld\n", kstate->challenge); 
#endif 
    k = htonl(kstate->challenge) ; 
    ident_writebase64((char *)&k,tmp,sizeof(k) ); 
    if ( global_auth_method->auth_flags != NULL ) { 
      sprintf(p,"%s, %s, %s",auth_data->auth_struct->auth_method,tmp,global_auth_method->auth_flags );
    } else { 
      sprintf(p,"%s, %s",auth_data->auth_struct->auth_method,tmp );
    }
    auth_data->auth_method_data = (void *) kstate ; 
    return ( p ) ; 
}

/*
 * Perform requester-side authentication protocol exchange
 * Returns an ident_auth error code.
 */
static int krb_requester_auth(auth_data)
 
  struct ident_auth_client_data *auth_data ;
{
    static KTEXT_ST authent;
    struct krb_state *kstate = (struct krb_state *) auth_data->auth_method_data;
    char tmp[IDENT_BUFFER_SIZE];
    char output[IDENT_BUFFER_SIZE];
    char *line_ptr, *tmp_ptr ; 
    char instance[INST_SZ]; 
    char uid[INST_SZ] ;
    char realm[REALM_SZ];
    int32_t from_addr ;
    IDENT_INT16 i, len , mutual_auth_flag, uid_len, requester_port,responder_port,munch ;
    int code,rcode;
    int maxbufsize;
    int ready ;     
    int timeout = IDENT_READ_TIMEOUT ; 

    while ( kstate->authstepno > -1 ) { 
      switch (kstate->authstepno++) {
      case 0:
	/* Get kerberos data*/
	if (strlen(auth_data->authenticate_data) > MIN_AUTH_LEN ) { 
	 rcode = sscanf(auth_data->authenticate_data,
		 " KERBEROS_V4 , %[A-Za-z0-9+/=]",tmp ) ;
	 if ( rcode != 1 ) { 
	   kstate->authstepno = -1;
	   return IDENT_INVALID_RESP_INFO;
	
	 }
	} else {  
	  kstate->authstepno = -1;
	  
	  return IDENT_INVALID_RESP_INFO;
	
	}
	srvtab =(char *) auth_data->auth_struct->auth_method_init ;
	from_addr = auth_data->remote_addr->s_addr ; 
	rcode = ident_decodebase64(tmp); 
	
	/*
	 * Client gave us ticket+authenticator
	 * reply with mutual auth
	 */
	munch = sizeof(IDENT_INT16) ; 
	memcpy((char *)&i,tmp, munch) ;
	authent.length = ntohs(i); 

	tmp_ptr = &(tmp[munch]) ; /* Hey, you'd do it this way too if you learned 
				C on a Cray YMP */ 
	memcpy(authent.dat, tmp_ptr , authent.length);
	authent.mbz = 0;
	strcpy(instance, "*");
	code = krb_rd_req(&authent, kstate->service, instance, from_addr,
			  &kstate->kdata, srvtab);
#ifdef DEBUG
	printf("krb_err_code %d , %s\n",code, krb_err_txt[code]); 
#endif
	if (code) {
	    kstate->authstepno = -1;
	    switch (code) {
	    case RD_AP_BADD:    /* This is common on dynamic ip clients */ 
	    case RD_AP_TIME:    
	    case RD_AP_EXP:
	      return IDENT_USER_CANT_AUTH; 
	      break; 
	    
	    default: 
	      return IDENT_INVALID_RESP_INFO;
	    }
	}
	if (kstate->kdata.checksum != kstate->challenge) {
	    kstate->authstepno = -1;
	    
	    return IDENT_MUTUAL_AUTH_FAIL;
	}

        /* Pull off the expiration time. */
        auth_data->expires = krb_life_to_time(kstate->kdata.time_sec,
                                              kstate->kdata.life);

	memcpy(kstate->session, kstate->kdata.session, sizeof(des_cblock));
	des_key_sched(kstate->session, kstate->schedule);

	/* Now get rest of info from tmp and check for integrity */ 
	tmp_ptr = &(tmp[authent.length + munch]) ; 
	memcpy(&i,tmp_ptr,munch); 
	len = ntohs(i) ; 
	tmp_ptr++ ; tmp_ptr++ ; 
	des_pcbc_encrypt((des_cblock *)tmp_ptr, (des_cblock *)output,len,kstate->schedule,&(kstate->session),0 ) ;
        
        memcpy(&mutual_auth_flag,output,munch) ;
	tmp_ptr = output+munch ; 
	
	memcpy(&responder_port,tmp_ptr,munch) ; 
	tmp_ptr += munch ; 
	memcpy(&requester_port,tmp_ptr,munch) ; 
	tmp_ptr += munch ; 
	memcpy(&uid_len,tmp_ptr,munch); 
	tmp_ptr += munch ; 
	mutual_auth_flag = ntohs(mutual_auth_flag); 
	requester_port = ntohs(requester_port) ; 
	responder_port = ntohs(responder_port) ; 
	uid_len = ntohs(uid_len); 
	
	if (( requester_port != auth_data->requester_port ) || 
	    ( responder_port != auth_data->responder_port )){ 
	    return IDENT_INVALID_RESP_INFO ;   
	} 
	tmp_ptr =&(output[4*munch]); 
	memcpy(uid,tmp_ptr,uid_len);  
	uid[uid_len] = '\0' ; 
	/* build user.inst@realm string and dump into user_ident */ 
	sprintf(auth_data->user_ident,"%s.%s@%s:%s",kstate->kdata.pname,
		kstate->kdata.pinst,kstate->kdata.prealm,uid ) ;

        /* Fill in principal if we should. */
        if ((krb_get_lrealm(realm, 1) == 0)
            && (strcmp(kstate->kdata.prealm, realm) == 0)
            && (kstate->kdata.pinst[0] == '\0'))
            strcpy(auth_data->user_principal,kstate->kdata.pname);

	if ( mutual_auth_flag & MUTUAL_AUTH_BIT ) {
	  return IDENT_AUTH_OKAY ; 
	} 
	
	
	break;
	
      case 1:
	/*
	 * Send back mutual auth reply.
	 */ 
 
	*(int *)output = htonl(kstate->challenge+1);
	tmp_ptr = output + sizeof(int32_t) ; 
	memset(tmp_ptr,'\0',sizeof(des_cblock)-sizeof(int32_t)); 
	des_ecb_encrypt(output, output, kstate->schedule, 1);
	ident_writebase64(output,tmp,sizeof(des_cblock) );
        

	sprintf(output,"%s, %s",auth_data->auth_struct->auth_method,tmp );
	do_out(output,auth_data); 
	
	return IDENT_AUTH_OKAY ; 

	break ; 
      default:
	kstate->authstepno = -1;
	return IDENT_SYSTEM_ERROR;
      }
    }
    return IDENT_AUTH_FAIL; /* make gcc -Wall happy */ 
}



/*
** Does trivial (i.e. passwd file based auth)
**
*/
static char *trivial_requester_start( auth_data)
struct ident_auth_data *auth_data; 
{ 
  return "\0" ; /* strlen NULL dies on SYSV */  

} 

static int trivial_requester_auth( auth_data)
struct ident_auth_data *auth_data; 
{ 
  return IDENT_AUTH_OKAY ; 
} 


static char *null_requester_start() { 
  return NULL ; 

}

static int     null_set_flag( char *flag, char *value ) { 
  return IDENT_AUTH_OKAY ; 
}

static int     null_get_flag( char *flag, char **value ) { 
  return IDENT_AUTH_OKAY ; 

} 

static int     null_requester_auth() { 

  return IDENT_AUTH_NOT_SUPPORTED ; 
}

static void     null_free_state() {
  return ; 
}
 






