/** @file
 * Glue routines for the pident kernel files.
 *
 * This file provides the glue routines needed to use pidentd 3.0.* k_files
 * with the S/Ident responder code.
 *
 * @author Booker C. Bense <bbense@stanford.edu>
 *
 * $Id: kernel_helper.c,v 1.2 2002/10/29 00:52:15 eagle Exp $
 */

#include "config.h" 

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>
#include <unistd.h>

#ifndef KERNEL_HELPER
# define KERNEL_HELPER
#endif 

#include "pidentd.h"  


/* osinfo_get is never actually used, since we don't use the
** ka_init function in the k_files.  Its just to make the 
** compiler happy. 
*/ 

char *osinfo_build = "the current os" ; 

char * 
osinfo_get( char *os_current) { 
  
  return NULL ; 

} 


struct kernel *
kernel_alloc(void)
{
    struct kernel *kp;
    
    kp = (struct kernel *) s_malloc(sizeof(struct kernel));
    
    
    kp->status = -1;
    kp->euid = NO_UID;
    kp->ruid = NO_UID;

    kp->pid = NO_PID;
    kp->cmd = NULL;
    kp->argv = NULL;
    
    return kp;
}


void
kernel_free(struct kernel *kp)
{
    s_free(kp->cmd);
    s_free(kp->argv);
    
    s_free( kp);
}




/*
** safeio.c - Signal/Async safe wrapper functions
**
** Copyright (c) 1997-1999 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

extern int debug_flag ; 

void
s_abort(void)
{

    abort();
}

/*
** A "safe" malloc, that always succeeds (or logs an
** error to syslog and the abort()'s.
**
** The buffer returned is zeroed out.
*/
void *
s_malloc(size_t size)
{
    void *p;

    p = (void *) malloc(size);
    if (p == NULL)
    {
	if (debug_flag)
	    fprintf(stderr, "s_malloc(%lu) failed - aborting\n",
		    (unsigned long) size);
	
	syslog(LOG_ERR, "malloc(%lu): %m", (unsigned long) size);
	s_abort();
    }

    memset(p, 0, size);
    return p;
}


void
s_free(void *p)
{
    if (p != NULL)
	free(p);
}


char *
s_strdup(const char *s)
{
    char *ns;
    size_t len;
    
    
    if (s == NULL)
	return NULL;

    len = strlen(s)+1;
    ns = (char *) malloc(len);
    if (ns == NULL)
    {
	syslog(LOG_ERR, "strdup(): malloc(%lu): %m", (unsigned long) len);
	s_abort();
    }

    memcpy(ns, s, len);
    return ns;
}

int
s_open(const char *path,
       int oflag,
       ...)
{
    int s;
    mode_t mode = 0;

    if (oflag & O_CREAT)
    {
	va_list ap;

	va_start(ap, oflag);
	/* FIXME: need to use widened form of mode_t here.  */
	mode = va_arg(ap, int);
	va_end(ap);
    }

    while ((s = open(path, oflag, mode)) < 0 && errno == EINTR)
	;
    
    if (s < 0 && (errno == EMFILE
		  || errno == ENFILE
		  || errno == ENOMEM 
#ifdef ENOSR
		  || errno == ENOSR
#endif
		  ))
    {
	/* Too many open files */
	
	syslog(LOG_WARNING, "s_open(\"%s\", 0%o): %m", path, oflag);
    }
    
    return s;
}




int
s_snprintf(char *buf,
	   size_t bufsize,
	   const char *format,
	   ...)
{
    va_list ap;
    int retcode;


    va_start(ap, format);

    if (bufsize < 1)
    {
        syslog(LOG_ERR, "s_snprintf(..., %d, ...): illegal bufsize",
	       bufsize);
	s_abort();
    }
    
    buf[bufsize-1] = '\0';
#ifdef HAVE_VSNPRINTF
    retcode = vsnprintf(buf, bufsize, format, ap);
#else
    retcode = vsprintf(buf, format, ap);
#endif
       
    if (retcode > 0 && (buf[bufsize-1] != '\0' ||
			retcode > bufsize-1))
    {
      syslog(LOG_ERR, "s_snprintf(..., %d, ...) = %d: buffer overrun\n",
	       bufsize, retcode);
	
	s_abort();
    }

    va_end(ap);

    return retcode;
}
